# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import UserDict
from livebridge.base.sources import BaseSource, PollingSource, StreamingSource
from livebridge.base.targets import BaseTarget
from livebridge.base.posts import BasePost
from livebridge.base.converters import BaseConverter
from livebridge.components import get_converter, get_db_client


class InvalidTargetResource(Exception):
    """This exception ca be raised, when a known resource target is invalid.
       This exception will prevent a retry of a failed action.
    """
    pass


class TargetResponse(UserDict):
    """Data container for returning the resource data from a service, awaits dictionary.

    Data returned from a BaseTarget derived target type has to be a TargetResponse.
    """
    pass


class ConversionResult(object):
    """Result object for conversions.

    :param content: content result of conversion
    :type string:
    :param images: list of local file paths of downloaded images, optional
    :type list:
    """
    def __init__(self, content, images=None):
        self.content = content
        self.images = images if images else []



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os.path
import uuid
import aiohttp


logger = logging.getLogger(__name__)


FILE_EXT = {
    "image/jpeg": ".jpg",
    "image/gif": ".gif",
    "image/png": ".png",
}


class BaseConverter(object):
    """Base class for converters.

    Concrete implementations have to implement public :func:`convert` method."""
    __module__ = "livebridge.base"

    source = ""
    target = ""

    def __init__(self):
        pass

    async def _download_image(self, data):
        basename = os.path.basename(data["media"])
        file_ext = os.path.splitext(basename)[1]
        filename = "{}-{}{}".format(
            str(uuid.uuid4())[:8],
            basename,
            FILE_EXT[data["mimetype"]] if not file_ext else ""
        )
        filepath = os.path.join("/tmp/", filename)
        with aiohttp.ClientSession() as session:
            async with session.get(data["href"]) as resp:
                test = await resp.read()
                with open(filepath, "wb") as file:
                    file.write(test)
        return filepath

    async def remove_images(self, images):
        try:
            for filepath in images:
                os.remove(filepath)
                logger.info("Removed image {}".format(filepath))
        except Exception as exc:
            logger.error("Error when deleting image.")
            logger.exception(exc)

    async def convert(self, post):
        """Convert incoming content of the incoming raw source post to a string suitable to the \
           target as content.

           :param post: raw source post
           :type dictionary: dictionary
           :returns: :class:`livebridge.base.ConversionResult` - result of the conversion."""
        raise NotImplementedError("Not implemented in converter.")



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from livebridge import config
from livebridge.storages import DynamoClient, SQLStorage, MongoStorage, DummyStorage


logger = logging.getLogger(__name__)


SOURCE_MAP = {}
CONVERTER_MAP = {}
POST_MAP = {}
TARGET_MAP = {}


def get_source(conf):
    client = None
    if SOURCE_MAP.get(conf.get("type")):
        source_cls = SOURCE_MAP[conf.get("type")]
        client = source_cls(config=conf)
    else:
        logger.warning("No source client found for {}.".format(conf))
    return client


def add_source(cls):
    SOURCE_MAP[cls.type] = cls


def get_converter(source, target):
    try:
        return CONVERTER_MAP[source][target]()
    except KeyError:
        logger.debug("no converter found for {} -> {}".format(source, target))
    return None


def add_converter(cls):
    CONVERTER_MAP.setdefault(cls.source, {})
    CONVERTER_MAP[cls.source][cls.target] = cls


def get_post(source, doc, content, images):
    return POST_MAP[source.type](doc, content=content, images=images)


def add_post(cls):
    POST_MAP[cls.source] = cls


def get_target(conf):
    client = None

    if TARGET_MAP.get(conf.get("type")):
        target_cls = TARGET_MAP[conf.get("type")]
        client = target_cls(config=conf)
    else:
        logger.warning("No target client found for {}.".format(conf))
    return client


def add_target(cls):
    TARGET_MAP[cls.type] = cls


def get_db_client(**kwargs):
    dsn = kwargs.get("dsn") or config.DB.get("dsn")
    if dsn:
        params = config.DB if not kwargs else kwargs
        if dsn.startswith("mongodb://"):
            return MongoStorage(**params)
        elif dsn.startswith("dummy://"):
            return DummyStorage(**params)
        else:
            return SQLStorage(**params)
    # default dynamodb
    params = config.AWS if not kwargs else kwargs
    return DynamoClient(**params)



# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
from datetime import datetime
from sqlalchemy_aio import ASYNCIO_STRATEGY
from sqlalchemy import create_engine, MetaData, Table, Column,\
                       Integer, String, Text, Boolean, DateTime
from sqlalchemy.schema import CreateTable
from sqlalchemy.sql import select
from livebridge.storages.base import BaseStorage


logger = logging.getLogger(__name__)

"""
http://docs.sqlalchemy.org/en/latest/core/engines.html
"""
class SQLStorage(BaseStorage):

    _instance = None
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(SQLStorage, cls).__new__(cls)
            logger.debug("SQL client: {}".format(cls._instance))
        return cls._instance

    def __init__(self, **kwargs):
        self.dsn = kwargs.get("dsn", None)
        self.table_name = kwargs.get("table_name")
        self.control_table_name = kwargs.get("control_table_name")

    @property
    async def db(self):
        if hasattr(self, "_engine") and self._engine:
            return self._engine
        logger.debug("Connecting to {}".format(self.dsn))
        self._engine = create_engine(self.dsn, strategy=ASYNCIO_STRATEGY)
        return self._engine

    def _get_table(self):
        return Table(self.table_name, MetaData(),
                     Column("id", Integer(), primary_key=True),
                     Column("target_id", String(150), index=True),
                     Column("post_id", String(150), index=True),
                     Column("source_id", String(150), index=True),
                     Column("sticky", Boolean()),
                     Column("text", Text()),
                     Column("created", DateTime()),
                     Column("updated", DateTime(), index=True),
                     Column("target_doc", Text()),
                    )

    def _get_control_table(self):
        return Table(self.control_table_name, MetaData(),
                     Column("id", Integer(), primary_key=True),
                     Column("type", String(150), index=True),
                     Column("data", Text()),
                     Column("updated", DateTime())
                    )

    async def setup(self):
        """Setting up SQL table, if it not exists."""
        try:
            engine = await self.db
            created = False
            if not await engine.has_table(self.table_name):
                # create table
                logger.info("Creating SQL table [{}]".format(self.table_name))
                items = self._get_table()
                await engine.execute(CreateTable(items))
                # create indeces
                conn = await engine.connect()
                await conn.execute(
                    "CREATE INDEX `lb_last_updated` ON `{}` (`source_id` DESC,`updated` DESC);".format(self.table_name))
                await conn.execute(
                    "CREATE INDEX `lb_post` ON `{}` (`target_id` DESC,`post_id` DESC);".format(self.table_name))
                await conn.close()
                created = True
            # create control table if not already created.
            if self.control_table_name and not await engine.has_table(self.control_table_name):
                # create table
                logger.info("Creating SQL control table [{}]".format(self.control_table_name))
                items = self._get_control_table()
                await engine.execute(CreateTable(items))
                created = True
            return created
        except Exception as exc:
            logger.error("[DB] Error when setting up SQL table: {}".format(exc))
        return False

    async def get_last_updated(self, source_id):
        try:
            db = await self.db
            table = self._get_table()
            result = await db.execute(
                            table.select().where(
                                table.c.source_id == source_id
                            ).order_by(
                                table.c.updated.desc()
                            ).limit(1))
            item = await result.first()
            tstamp = item["updated"] if item else None
            return tstamp
        except Exception as exc:
            logger.error("[DB] Error when querying for last updated item on {}".format(source_id))
            logger.exception(exc)
        return None

    async def get_known_posts(self, source_id, post_ids):
        results = []
        try:
            db = await self.db
            table = self._get_table()
            sql = select(columns=[table.c.post_id]).where(table.c.source_id == source_id).where(
                table.c.post_id.in_(post_ids))
            db_res = await db.execute(sql)
            for row in await db_res.fetchall():
                results.append(row[0])
        except Exception as exc:
            logger.error("[DB] Error when querying for posts {}".format(post_ids))
            logger.exception(exc)
        return results

    async def get_post(self, target_id, post_id):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.select().where(table.c.target_id == target_id).where(table.c.post_id == post_id).limit(1)
            result = await db.execute(sql)
            item = await result.first()
            if item:
                item = dict(item)
                item["target_doc"] = json.loads(item["target_doc"]) if item.get("target_doc") != "" else {}
            return item
        except Exception as exc:
            logger.error("[DB] Error when querying for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return None

    async def insert_post(self, **kwargs):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.insert().values(
                target_id=kwargs.get("target_id"),
                post_id=kwargs.get("post_id"),
                source_id=kwargs.get("source_id"),
                text=kwargs.get("text"),
                sticky=int(kwargs.get("sticky", "0")),
                created=kwargs.get("created"),
                updated=kwargs.get("updated"),
                target_doc=json.dumps(kwargs.get("target_doc")) if kwargs.get("target_doc") else ""
            )
            await db.execute(sql)
            logger.info("[DB] Post {} {} was saved!".format(kwargs["source_id"], kwargs["post_id"]))
            return True
        except Exception as exc:
            logger.error("[DB] Error when saving {}".format(kwargs))
            logger.error(exc)
        return False

    async def update_post(self, **kwargs):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.update().where(
                table.c.target_id == kwargs.get("target_id")
                ).where(
                    table.c.post_id == kwargs.get("post_id")
                ).values(
                    target_id=kwargs.get("target_id"),
                    post_id=kwargs.get("post_id"),
                    source_id=kwargs.get("source_id"),
                    text=kwargs.get("text"),
                    sticky=int(kwargs.get("sticky", "0")),
                    created=kwargs.get("created"),
                    updated=kwargs.get("updated"),
                    target_doc=json.dumps(kwargs.get("target_doc")) if kwargs.get("target_doc") else ""
                )
            await db.execute(sql)
            logger.info("[DB] Post {} {} was updated!".format(kwargs.get("post_id"), kwargs.get("target_id")))
            return True
        except Exception as exc:
            logger.error("[DB] Error when updating for a post [{}] on {}".format(
                kwargs.get("post_id"), kwargs.get("target_id")))
            logger.error(exc)
        return False

    async def delete_post(self, target_id, post_id):
        try:
            db = await self.db
            table = self._get_table()
            sql = table.delete().where(table.c.target_id == target_id).where(table.c.post_id == post_id)
            await db.execute(sql)
            logger.info("[DB] Post {} {} was deleted!".format(target_id, post_id))
            return True
        except Exception as exc:
            logger.error("[DB] Error when deleting for a post [{}] on {}".format(post_id, target_id))
            logger.error(exc)
        return False

    async def get_control(self, updated=None):
        try:
            db = await self.db
            table = self._get_control_table()
            sql = table.select().where(table.c.type == "control")
            if updated: # check for updated timestamp
                sql = sql.where(table.c.updated != updated)
            sql = sql.limit(1)
            result = await db.execute(sql)
            item = await result.first()
            if item:
                item = dict(item)
                item["data"] = json.loads(item["data"]) if item.get("data") != "" else {}
                return item
        except Exception as exc:
            logger.error("[DB] Error when querying for a control data on {}".format(self.control_table_name))
            logger.error(exc)
        return False

    async def save_control(self, data):
        try:
            db = await self.db
            table = self._get_control_table()
            existing = await self.get_control()
            updated=datetime.now()
            if existing:
                # update
                sql = table.update().where(table.c.type == "control").values(
                    data=json.dumps(data),
                    updated=updated
                )
            else:
                # create
                sql = table.insert().values(
                    type="control",
                    data=json.dumps(data),
                    updated=updated
                )
            await db.execute(sql)
            logger.info("[DB] Control data was saved.")
            return True
        except Exception as exc:
            logger.error("[DB] Error when saving control data on {}".format(self.control_table_name))
            logger.error(exc)
        return False



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from collections import UserDict


log = logging.getLogger(__name__)


class LiveBridge(UserDict):
    """Livebridge object.

    :param loop: loop to use.
    :type :py:class:`asyncio.BaseEventLoop`
    :param controller: Livebridge controller
    :type :class:`livebridge.controller.Controller`
    """
    def __init__(self, loop, controller):
        self.loop = loop
        self.controller = controller

    # clean up at tear down
    async def finish(self, tasks):
        log.info("Finishing ...")
        for task in tasks:
            log.debug("... {}".format(task))
            task.cancel()
        log.info("Bye!")

    def shutdown(self):
        self.loop.run_until_complete(self.controller.clean_shutdown())
        self.loop.run_until_complete(self.finish(self.controller.tasked))
        if self.loop._default_executor:
            self.loop._default_executor.shutdown(wait=True)





# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import asyncio
import functools
import logging
import signal
import livebridge.logger
from argparse import Namespace
from livebridge import config, LiveBridge
from livebridge.controller import Controller
from livebridge.components import get_db_client
from livebridge.web import WebApi
from livebridge.loader import load_extensions


def read_args(**kwargs):
    """Read controlfile parameter."""
    if kwargs.get("control"):
        args = Namespace(control=kwargs["control"])
    elif config.CONTROLFILE:
        args = Namespace(control=config.CONTROLFILE)
    elif config.DB.get("control_table_name"):
        args = Namespace(control="sql")
    elif config.AWS.get("control_table_name"):
        args = Namespace(control="dynamodb")
    else:
        # read cli args
        parser = argparse.ArgumentParser()
        parser.add_argument("--control", required=True, help="Control file, can be path.")
        args = parser.parse_args()
    return args


def main(**kwargs):
    # disable bot logging
    logging.getLogger('botocore').setLevel(logging.ERROR)
    logging.getLogger('websockets').setLevel(logging.INFO)

    # read args
    args = read_args(**kwargs)

    # initialize loop
    loop = asyncio.get_event_loop() if not kwargs.get("loop") else kwargs["loop"]

    # load extensions
    load_extensions()

    # setup db table
    db_connector = get_db_client()
    loop.run_until_complete(db_connector.setup())

    # controller manages the tasks / data
    controller = Controller(config=config, control_file=args.control)
    asyncio.ensure_future(controller.run())

    # start http api
    if config.WEB.get("host") and config.WEB.get("port"):
        server = WebApi(config=config.WEB, controller=controller, loop=loop)

    # add signal handler
    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(getattr(signal, signame), functools.partial(loop.stop))

    livebridge = LiveBridge(loop=loop, controller=controller)

    if kwargs.get("loop"):
        return livebridge

    # run
    try:
        loop.run_forever()
    finally:
        if config.WEB.get("host") and config.WEB.get("port"):
            server.shutdown()
        livebridge.shutdown()
        loop.stop()
        loop.close()



# -*- coding: utf-8 -*-
#
# Copyright 2016, 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import aiobotocore
import json
import logging
import os
import os.path
import yaml
from botocore.client import Config
from botocore.exceptions import ClientError
from livebridge.config import AWS
from livebridge.controldata.base import BaseControl

logger = logging.getLogger(__name__)

class ControlFile(BaseControl):

    def __init__(self):
        self._sqs_client = None
        self._s3_client = None
        self.config = AWS
        self._updated_local = None

    def __del__(self):
        if self._sqs_client:
            self._sqs_client.close()

        if self._s3_client:
            self._s3_client.close()

    @property
    async def sqs_client(self):
        if self._sqs_client:
            return self._sqs_client
        session = aiobotocore.get_session()
        self._sqs_client = session.create_client('sqs',
             region_name=self.config["region"],
             aws_secret_access_key=self.config["secret_key"] or None,
             aws_access_key_id=self.config["access_key"] or None)

        await self._purge_sqs_queue()

        return self._sqs_client

    @property
    def s3_client(self):
        if self._s3_client:
            return self._s3_client
        session = aiobotocore.get_session()
        self._s3_client = session.create_client('s3',
             region_name=self.config["region"],
             aws_secret_access_key=self.config["secret_key"] or None,
             aws_access_key_id=self.config["access_key"] or None)
        return self._s3_client

    async def _purge_sqs_queue(self):
        # purge queue before starting watching
        try:
            await self._sqs_client.purge_queue(
                QueueUrl=self.config["sqs_s3_queue"]
            )
            logger.info("Purged SQS queue {}".format(self.config["sqs_s3_queue"]))
        except ClientError as exc:
            logger.warning("Purging SQS queue failed with: {}".format(exc))

    async def check_control_change(self, control_path=None):
        if not self.config.get("sqs_s3_queue", False):
            return await self._check_local_changes(control_path)
        else:
            return await self._check_s3_changes()

    async def _check_local_changes(self, control_path):
        is_changed = False
        try:
            last_updated = os.stat(control_path).st_mtime
            if last_updated != self._updated_local:
                is_changed = True
            self._updated_local = last_updated
        except Exception as exc:
            logger.error("Error fetching last updated time of local control file: {}".format(exc))
        return is_changed

    async def _check_s3_changes(self):
        client = await self.sqs_client
        # check for update events
        try:
            response = await client.receive_message(
                QueueUrl=self.config["sqs_s3_queue"]
            )
            for msg in response.get("Messages", []):
                logger.debug("SQS {}".format(msg.get("MessageId")))
                body = msg.get("Body")
                data = json.loads(body) if body else None
                await client.delete_message(
                    QueueUrl=self.config["sqs_s3_queue"],
                    ReceiptHandle=msg.get("ReceiptHandle")
                )
                if data:
                    for rec in data.get("Records", []):
                        logger.debug("EVENT: {} {}".format(
                            rec.get("s3", {}).get("object", {}).get("key"), rec.get("eventName")))
                        return True
        except Exception as exc:
            logger.error("Error fetching SQS messages with: {}".format(exc))
        return False

    async def load(self, path, *, resolve_auth=False):
        if not path.startswith("s3://"):
            body = self._load_from_file(path)
        else:
            body = await self._load_from_s3(path)
        return yaml.load(body)

    def _load_from_file(self, path):
        logger.info("Loading control file from disk: {}".format(path))
        if not os.path.exists(path):
            raise IOError("Path for control file not found.")

        file = open(path, "r")
        body = file.read()
        self._updated_local = os.stat(path).st_mtime
        file.close()

        return body

    async def _save_to_file(self, path, data):
        logger.info("Saving control file to disk.")
        directory = os.path.dirname(os.path.abspath(path))
        if not os.access(directory, os.W_OK):
            raise IOError("Path for control file not writable: {}".format(path))

        file = open(path, "w+")
        body = file.write(data)
        file.close()

        return True

    async def _load_from_s3(self, url):
        bucket, key = url.split('/', 2)[-1].split('/', 1)
        logger.info("Loading control file from s3: {} - {}".format(bucket, key))
        control_file = await self.s3_client.get_object(Bucket=bucket, Key=key)
        control_data = await control_file["Body"].read()
        return control_data

    async def _save_to_s3(self, path, data):
        bucket, key = path.split('/', 2)[-1].split('/', 1)
        logger.info("Saving control file to s3: {} - {}".format(bucket, key))
        response = await self.s3_client.put_object(Body=data, Bucket=bucket, Key=key)
        return True

    async def save(self, path, data):
        res = False
        yaml_data = yaml.dump(data, indent=4, default_flow_style=False)
        if not path.startswith("s3://"):
            res = await self._save_to_file(path, yaml_data)
        else:
            res = await self._save_to_s3(path, yaml_data)
        return res



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
from livebridge.components import get_target
from livebridge.bridge import LiveBridge
from livebridge.controldata import ControlData

logger = logging.getLogger(__name__)

class Controller(object):

    def __init__(self, config, control_file):
        self.config = config
        self.control_file = control_file
        self.poll_interval = config.POLL_INTERVAL or 60
        self.check_control_interval = config.POLL_CONTROL_INTERVAL or 60
        self.read_control = False
        self.tasked = []
        self.sleep_tasks = []
        self.bridges = {}
        self.retry_run_interval = 30
        self.control_data = None # access to data from control file
        self.shutdown = False

    async def clean_shutdown(self):
        logger.info("Requesting proper shutdown of tasks.")
        self.shutdown = True
        await self.stop_bridges()
        while len(self.bridges) > 0:
            logger.debug("Running bridges left: {}".format(len(self.bridges)))
            await asyncio.sleep(1)

    async def stop_bridges(self):
        """Stop all sleep tasks to allow bridges to end."""
        for task in self.sleep_tasks:
            task.cancel()
        for bridge in self.bridges:
            bridge.stop()

    async def check_control_change(self):
        # check for update events
        logger.info("Starting watching for control data changes.")
        while True and self.shutdown != True:
            is_changed = await self.control_data.check_control_change(self.control_file)
            if is_changed == True:
                if self.bridges:
                    # running bridges
                    self.read_control = True
                    await self.stop_bridges()
                else:
                    logger.info("No running bridges found, start with new control data.")
                    asyncio.ensure_future(self.run())
                return True
            await self.sleep(self.check_control_interval)

    def append_bridge(self, config_data):
        bridge = LiveBridge(config_data)
        for tconf in config_data.get("targets", []):
            target_client = get_target(tconf)
            bridge.add_target(target_client)
        # start specific source
        if bridge.source.mode == "streaming":
            self.bridges[bridge] = self.run_stream(bridge=bridge)
        elif bridge.source.mode == "polling":
            # custom source polling intervall set in control file?
            poll_interval = config_data["poll_interval"] if config_data.get("poll_interval") else self.poll_interval
            self.bridges[bridge] = self.run_poller(bridge=bridge, interval=poll_interval)

    def remove_bridge(self, bridge):
        logger.info("ENDING: {}".format(bridge))
        del self.bridges[bridge]
        if self.read_control and not self.bridges:
            logger.info("RESTART BRIDGES")
            asyncio.ensure_future(self.run())

    async def save_control_data(self, doc):
        control_data = ControlData(config=self.config)
        return await control_data.save(self.control_file, doc)

    async def load_control_data(self, resolve_auth=True):
        control_data = ControlData(config=self.config)
        await control_data.load(self.control_file, resolve_auth=resolve_auth)
        return control_data

    async def retry_run(self):
        logger.info("Will retry loading control file in 30 seconds.")
        await asyncio.sleep(self.retry_run_interval)
        asyncio.ensure_future(self.run())

    async def run(self):
        self.read_control = False
        loaded = False
        control_data = None
        try:
            control_data = await self.load_control_data()
            loaded = True
        except Exception as exc:
            logger.error("Error when reading control file.")
            logger.error(exc)
            logger.info("Will try to reuse existing control data.")

        # set new control data
        if loaded:
            logger.info("Using fetched control data.")
            self.control_data = control_data

        # (re)start tasks or retry fetching control file
        if self.control_data:
            self.start_tasks()
        else:
            self.tasked.append(asyncio.Task(self.retry_run()))

    def start_tasks(self):
        # append content bridges
        for bridge_config in self.control_data.list_bridges():
            self.append_bridge(bridge_config)

        # create futures for bridges
        for bridge in self.bridges:
            self.tasked.append(asyncio.Task(self.bridges[bridge]))

        # listen to s3 control changes
        self.tasked.append(asyncio.Task(self.check_control_change()))

    async def run_stream(self, *, bridge):
        await bridge.listen_ws()
        while True and self.read_control != True and self.shutdown != True:
            # wait for shutdown
            await self.sleep(4)

        # stop bridge
        try:
            await bridge.source.stop()
        except Exception as exc:
            logger.error("Error when stopping stream: {}".format(exc))

        self.remove_bridge(bridge)

    async def sleep(self, seconds):
        if self.read_control is True or self.shutdown is True:
            # if shutdown or restart is requested, don't fall asleep again
            return True

        try:
            task = asyncio.ensure_future(asyncio.sleep(seconds))
            self.sleep_tasks.append(task)
            await task
        except asyncio.CancelledError:
            pass
        finally:
            self.sleep_tasks.remove(task)
        return True

    async def run_poller(self, *, bridge, interval=180):
        # initialize liveblogs to watch
        while True and self.read_control != True and self.shutdown != True:
            #logger.debug("Checked new posts for {} on {}".format(bridge.source_id, bridge.endpoint))
            await bridge.check_posts()
            await self.sleep(interval)

        self.remove_bridge(bridge)
        return



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys

from livebridge.config import LOGFILE, LOGLEVEL

logger = logging.getLogger()
logger.setLevel(getattr(logging, LOGLEVEL))
formatter = logging.Formatter('%(asctime)s [%(name)s] %(levelname)s %(message)s')

ch = logging.StreamHandler(sys.stdout)
ch.setFormatter(formatter)
logger.addHandler(ch)

if LOGFILE:
    fh = logging.FileHandler(LOGFILE)
    fh.setFormatter(formatter)
    logger.addHandler(fh)



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from livebridge.components import get_db_client


logger = logging.getLogger(__name__)


class BaseSource(object):
    """Base class for sources."""
    __module__ = "livebridge.base"

    type = ""
    mode = ""

    def __init__(self, *, config={}, **kwargs):
        """Base constructor for sources.

        :param config: Configuration passed from the control file.
        """
        pass

    @property
    def _db(self):
        """Database client for accessing storage.

           :returns: :class:`livebridge.storages.base.BaseStorage` """
        if not hasattr(self, "_db_client") or getattr(self, "_db_client") is None:
            self._db_client = get_db_client()
        return self._db_client

    async def filter_new_posts(self, source_id, post_ids):
        """Filters ist of post_id for new ones.

        :param source_id: id of the source
        :type string:
        :param post_ids: list of post ids
        :type list:
        :returns: list of unknown post ids."""
        new_ids = []
        try:
            db_client = self._db
            posts_in_db = await db_client.get_known_posts(source_id, post_ids)
            new_ids = [p for p in post_ids if p not in posts_in_db]
        except Exception as exc:
            logger.error("Error when filtering for new posts {} {}".format(source_id, post_ids))
            logger.exception(exc)
        return new_ids

    async def get_last_updated(self, source_id):
        """Returns latest update-timestamp from storage for source.

        :param source_id: id of the source (source_id, ticker_id, blog_id pp)
        :type string:
        :returns: :py:class:`datetime.datetime` object of latest update datetime in db."""
        last_updated = await self._db.get_last_updated(source_id)
        logger.info("LAST UPDATED: {} {}".format(last_updated, self))
        return last_updated


class PollingSource(BaseSource):
    """Base class for sources which are getting polled. Any custom adapter source, which \
       should get polled, has to be inherited from this base class."""

    mode = "polling"

    async def poll(self):
        """Method has to be implemented by the concrete inherited source class.

        :func:`poll` gets called by the interval defined by environment var *POLLING_INTERVALL*.

        The inheriting class has to implement the actual poll request for the source in this method.

        :return: list of new posts"""
        raise NotImplementedError("Method 'poll' not implemented.")


class StreamingSource(BaseSource):
    """Base class for streaming sources. Any custom adapter source, which is using a websocket, SSE or\
       any other stream as source has to be inherited from this base class."""

    mode = "streaming"

    async def listen(self, callback):
        """Method has to be implemented by the concrete inherited source class.

        A websocket connection has to be opened and given *callback* method has to be
        called with the new post as argument.

        :param callback: Callback method which has to be called with list of new posts.
        :return: True"""
        raise NotImplementedError("Method 'listen' not implemented.")

    async def stop(self):
        """Method has to be implemented by the concrete inherited source class.

           By calling this method, the websocket-connection has to be stopped.

           :return: True"""
        raise NotImplementedError("Method 'stop' not implemented.")



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class BaseStorage(object):
    """Base class for storage clients.

    Every concrete implementation of a storage class, requires to implement following methods:

    * :func:`db` - property holds underlying databse client.
    * :func:`setup` - Method for setting up storage, creating table pp.
    * :func:`insert_post` - insert single post into storage.
    * :func:`get_post` - returns single post from storage.
    * :func:`update_post` - updates single post in storage.
    * :func:`delete_post` - deletes single post in storage.
    * :func:`get_last_updated` - returns latest updated-timestamp or source from storage.
    """

    @property
    async def db(self):
        """Property holds underlying database client."""
        raise NotImplementedError()

    def setup(self):
        """Method for setting up storage, creating table pp."""
        raise NotImplementedError()

    async def insert_post(self, **kwargs):
        """Insert single post into storage."""
        raise NotImplementedError()

    async def get_post(self, target_id, post_id):
        """Returns single post from storage.

        :param target_id: id of the target
        :type string:
        :param post_id: id of post
        :type string:"""
        raise NotImplementedError()

    async def update_post(self, **kwargs):
        """Updates single post in storage."""
        raise NotImplementedError()

    async def delete_post(self, target_id, post_id):
        """Deletes single post from storage.

        :param target_id: id of the target
        :type string:
        :param post_id: id of post to delete
        :type string:"""
        raise NotImplementedError()

    async def get_last_updated(self, source_id):
        """Returns latest updated-timestamp of source from storage.

        :param source_id: id of the source
        :type string:"""
        raise NotImplementedError()

    async def get_known_posts(self, source_id, post_ids):
        """Return a list of known post_id of a source for a given list of post ids.

        :param source_id: id of the source
        :type string:
        :param post_ids: list of post ids
        :type list:
        :returns: - list of dictionaries."""
        raise NotImplementedError()

    async def get_control(self):
        """Method for retrieving of control data form storage.

        :returns: - dictionary"""
        raise NotImplementedError()

    async def save_control(self, data):
        """Method for saving control data in storage.

        :param data: control_data
        :type dict:
        :returns: - boolean"""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from setuptools import setup, find_packages
import sys, os

version = '0.21.0'

setup(name='livebridge',
      version=version,
      description="Keep content in-sync across services. Or simply syndicate content to multiple services. Based on asyncio.",
      long_description="""\
See https://github.com/dpa-newslab/livebridge for more infos.
""",
      classifiers=[
        "Programming Language :: Python :: 3.5",
        "Topic :: Communications :: Chat", 
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Other Audience",
        "Operating System :: POSIX :: Linux",
        "Environment :: Console",
        ], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords=['liveticker','liveblog','syndication', "async", "asyncio"],
      author='dpa-infocom GmbH',
      maintainer='Martin Borho',
      maintainer_email='martin@borho.net',
      url='https://github.com/dpa-newslab/livebridge',
      license='Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)',
      packages=find_packages(exclude=['tests', 'htmlcov', 'dist',]),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'aiobotocore==0.3.3',
        'aiohttp==2.1.0',
        'asynctest==0.10.0',
        'bleach==2.0.0',
        'PyYAML==3.12',
        'websockets==3.3',
        'sqlalchemy-aio==0.11.0',
        'python-dateutil==2.6.0',
        'dsnparse==0.1.4',
        'motor==1.1',
      ],
      entry_points="""
      [console_scripts]
      livebridge = livebridge.run:main
      """,
      )




# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class BasePost(object):
    """Base class for posts.

    Every concrete implementation of this base class have to implement the \
    poperties defined here. This is needed, so the specific targets do not have \
    to differentiate by the post type, but rather have a unified way to access \
    common properties from the resource, regardless of their source."""
    __module__ = "livebridge.base"

    source = ""

    def __init__(self, data, *, content="", images=[]):
        """Base constructor for targets.

        :param data: - dictionary of source post data
        :param content: - string of convertered post content, optional
        :param images: - list of iage_paths, optional"""
        self.data = data
        self.content = content
        self.images = images
        self._existing = None
        self._target_id = None
        self._target_doc = None

    @property
    def id(self):
        """Returns the id of the resource at its source."""
        raise NotImplementedError("Property not implemented.")

    @property
    def source_id(self):
        """Returns the id of the source. For example the id of the blog polled."""
        raise NotImplementedError("Property not implemented.")

    @property
    def created(self):
        """Returns created datetime of the resource, has to be a :py:class:`datetime.datetime` object."""
        raise NotImplementedError("Property not implemented.")

    @property
    def updated(self):
        """Returns updated datetime of the resource, has to be a :py:class:`datetime.datetime` object."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_update(self):
        """Returns boolean if resource is updated or not."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_deleted(self):
        """Returns boolean if resource is deleted or not."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_sticky(self):
        """Returns boolean if resource is sticky or not."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_submitted(self):
        """Returns boolean if resource is in submitted state."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_draft(self):
        """Returns boolean if resource is in drafted state."""
        raise NotImplementedError("Property not implemented.")

    @property
    def is_known(self):
        """Returns boolean if resource is already known to livebridge."""
        return bool(self._existing)

    def get_action(self):
        """Returns type of action which has to be handled by the target.
           Can be either **create**, **update**, **delete** or **ignore**."""
        raise NotImplementedError("Method not implemented.")

    @property
    def target_doc(self):
        """Returns resource doc as at the target, when the posting was already created \
           at the target. This property normally contains the **target_doc** data from \
           the livebrigde storage item, saved in a syndication earlier.

        :returns: dict"""
        if not hasattr(self, "_target_doc") or not self._target_doc:
            if self._existing:
                self._target_doc = self._existing.get("target_doc", {})
        return self._target_doc

    @target_doc.setter
    def target_doc(self, value):
        """Setter method for **target_doc** property."""
        self._target_doc = value

    @property
    def target_id(self):
        """Returns the id the target, to which this post has to be syndicated.

        :returns: string"""
        # already set?
        if self._target_id:
            return self._target_id
        # post already exists?
        if self._existing:
            self._target_id = self._existing.get("target_id")
        return self._target_id

    def set_existing(self, existing):
        """Takes existing doc at target.

        :param existing: - dict, resource doc at target."""
        self._existing = existing

    def get_existing(self):
        """Returns existing resource at target.

        :returns: - dict of resource"""
        return self._existing



# -*- coding: utf-8 -*-
#
# Copyright 2016 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from livebridge.components import get_converter, get_db_client


logger = logging.getLogger(__name__)


class BaseTarget(object):
    """Base class for targets.

    Every concrete implementation of a target, requires to implement following methods:

    * :func:`post_item` - creates new resource at target
    * :func:`delete_item` - deletes resoruce at target
    * :func:`update_item` - updates resource at target
    * :func:`handle_extras` - handle extra actions, for example set resource to *sticky* \
       at the target.

    If a method above is not needed, for example when no update is possible at the target, \
    use :func:`livebridge.posts.base.BasePost.get_action` to **ignore** this action."""
    __module__ = "livebridge.base"

    type = "base"

    def __init__(self, *, config={}, **kwargs):
        """Base constructor for targets.

        :param config: Configuration passed from the control file.
        """
        self.target_id = None

    @property
    def _db(self):
        if not hasattr(self, "_db_client") or getattr(self, "_db_client") is None:
            self._db_client = get_db_client()
        return self._db_client

    def post_item(self, post):
        """Creates new resource at target.

        :param post: - Object of type :class:`livebridge.posts.base.BasePost`
        :returns: - :class:`livebridge.base.TargetResponse` of the newly created resource returned \
        from service, empty if failed."""
        raise NotImplementedError()

    def delete_item(self, post):
        """Deletes an existing resource at target.

        :param post: - Object of type :class:`livebridge.posts.base.BasePost`
        :returns: - :class:`livebridge.base.TargetResponse` of the deleted resource \
        returned from service, empty or False if failed."""
        raise NotImplementedError()

    def update_item(self, post):
        """Updates existing resource resource at target.

        :param post: - Object of type :class:`livebridge.posts.base.BasePost`
        :returns: - :class:`livebridge.base.TargetResponse` of the updated resource \
        returned from service, empty or False if failed."""
        raise NotImplementedError()

    def handle_extras(self, post):
        """Handle extra actions, for example set resource to *sticky* at the target. Return None \
        if method is unneeded.

        :param post: - Object of type :class:`livebridge.posts.base.BasePost`
        :returns: - :class:`livebridge.base.TargetResponse` of the updated resource returned \
        from service, empty or None if nothing has changed at resource."""
        raise NotImplementedError()

    async def _handle_new(self, post):
        new_doc = await self.post_item(post)
        logger.debug("Target CREATE: {}".format(new_doc))
        if not new_doc:
            logger.error("Post {} wasn't saved in {}".format(post.id, self.target_id))
            return ""
        return new_doc

    async def _handle_delete(self, post):
        del_res = await self.delete_item(post)
        logger.debug("Target DELETE: {}".format(del_res))
        if not del_res:
            logger.error("Deleting post failed: [{}] on {}".format(post.id, post.target_id))
            return False
        await self._db.delete_post(
            post.target_id,
            post.id
        )
        logger.info("Deleted post: [{}] on {}".format(post.id, post.target_id))
        return True

    async def _handle_update(self, post):
        update_res = await self.update_item(post)
        logger.debug("Target UPDATE: {}".format(update_res))
        if not update_res:
            logger.error("Target Update failed {} / {}".format(post.id, post.content))

        logger.info("Updated post: [{}] on {}".format(post.id, post.target_id))
        return update_res

    def _get_converter(self, post):
        return get_converter(post.source, self.type)

    async def handle_post(self, post):
        # convert
        converter = self._get_converter(post)
        if converter:
            # convert from source to target
            conversion = await converter.convert(post.data)
            post.content = conversion.content
            post.images = conversion.images
            logger.debug("CONVERSION RESULTS: {}".format(post.content))

            if not post.content and not post.is_deleted:
                logger.warning("Empty text, post got ignored.")
                return None

        post.set_existing(await self._db.get_post(self.target_id, post.id))
        action = post.get_action()
        logger.info("POST ACTION: {} - {} - {}".format(action, self.target_id, post.id))
        if action == "ignore":
            return None
        elif action == "create":
            post.target_doc = await self._handle_new(post)
        elif action == "update":
            post.target_doc = await self._handle_update(post)
        elif action == "delete":
            await self._handle_delete(post)
            return None

        # handle extra traits of target
        extra_doc = await self.handle_extras(post)
        if extra_doc:
            # use updated doc from target
            post.target_doc = extra_doc

        if post:
            # save new doc
            put_params = {
                "target_id": self.target_id,
                "post_id": post.id,
                "source_id": post.source_id,
                "text": str(post.content),
                "sticky": post.is_sticky,
                "created": post.created,
                "updated": post.updated,
                "target_doc": post.target_doc.data,
            }
            if action == "create":
                await self._db.insert_post(**put_params)
            elif action == "update":
                await self._db.update_post(**put_params)

        # clean up converter images
        if converter:
            await converter.remove_images(post.images)



# -*- coding: utf-8 -*-
#
# Copyright 2017 dpa-infocom GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from livebridge.controldata.base import BaseControl
from livebridge.components import get_db_client

logger = logging.getLogger(__name__)

class StorageControl(BaseControl):

    def __init__(self):
        self._db_client = None
        self._updated = None

    @property
    async def db_client(self):
        if self._db_client:
            return self._db_client
        self._db_client = get_db_client()
        return self._db_client

    async def  _load_control_data(self, updated=None):
        db_client = await self.db_client
        return await db_client.get_control(updated=updated)

    async def check_control_change(self, *, control_path=None):
        #if not self._updated: # do not check without prior data
        #    return False

        try:
            control_data = await self._load_control_data(updated=self._updated)
            if control_data:
                return True
        except Exception as exc:
            logger.error("Error checking sql control data change with: {}".format(exc))
        return False

    async def load(self, path):
        control_data = await self._load_control_data()
        if control_data and control_data.get("updated"):
            self._updated = control_data["updated"]
            return control_data.get("data")
        return {}

    async def save(self, path, data):
        db_client = await self.db_client
        return await db_client.save_control(data)
